@@ -1,5 +1,6 @@
 """Functions for working with features in a raster dataset."""
 
+from contextlib import ExitStack
 import logging
 import math
 import os
@@ -11,10 +12,10 @@ import rasterio
 from rasterio import warp
 from rasterio._base import DatasetBase
 from rasterio._features import _shapes, _sieve, _rasterize, _bounds
-from rasterio.dtypes import validate_dtype, can_cast_dtype, get_minimum_dtype, _getnpdtype
 from rasterio.enums import MergeAlg
 from rasterio.env import ensure_env, GDALVersion
 from rasterio.errors import ShapeSkipWarning
+from rasterio.io import DatasetWriter
 from rasterio.rio.helpers import coords
 from rasterio.transform import Affine
 from rasterio.transform import IDENTITY, guard_transform
@@ -39,23 +40,24 @@ def geometry_mask(
     ----------
     geometries : iterable over geometries (GeoJSON-like objects)
     out_shape : tuple or list
-        Shape of output numpy ndarray.
+        Shape of output :class:`numpy.ndarray`.
     transform : Affine transformation object
         Transformation from pixel coordinates of `source` to the
         coordinate system of the input `shapes`. See the `transform`
         property of dataset objects.
     all_touched : boolean, optional
         If True, all pixels touched by geometries will be burned in.  If
-        false, only pixels whose center is within the polygon or that
+        False, only pixels whose center is within the polygon or that
         are selected by Bresenham's line algorithm will be burned in.
+        False by default
     invert: boolean, optional
         If True, mask will be True for pixels that overlap shapes.
         False by default.
 
     Returns
     -------
-    numpy ndarray of type 'bool'
-        Result
+    numpy.ndarray :
+        Type is :class:`numpy.bool_`
 
     Notes
     -----
@@ -79,15 +81,15 @@ def shapes(source, mask=None, connectivity=4, transform=IDENTITY):
 
     Parameters
     ----------
-    source : array, dataset object, Band, or tuple(dataset, bidx)
-        Data type must be one of rasterio.int16, rasterio.int32,
-        rasterio.uint8, rasterio.uint16, or rasterio.float32.
-    mask : numpy ndarray or rasterio Band object, optional
+    source : numpy.ndarray, dataset object, Band, or tuple(dataset, bidx)
+        Data type must be one of rasterio.int8, rasterio.int16, rasterio.int32,
+        rasterio.uint8, rasterio.uint16, rasterio.float32, or rasterio.float64.
+    mask : numpy.ndarray or rasterio Band object, optional
         Must evaluate to bool (rasterio.bool_ or rasterio.uint8). Values
         of False or 0 will be excluded from feature generation.  Note
         well that this is the inverse sense from Numpy's, where a mask
         value of True indicates invalid data in an array. If `source` is
-        a Numpy masked array and `mask` is None, the source's mask will
+        a :class:`numpy.ma.MaskedArray` and `mask` is None, the source's mask will
         be inverted and used in place of `mask`.
     connectivity : int, optional
         Use 4 or 8 pixel connectivity for grouping pixels into features
@@ -123,8 +125,7 @@ def shapes(source, mask=None, connectivity=4, transform=IDENTITY):
         source = source.data
 
     transform = guard_transform(transform)
-    for s, v in _shapes(source, mask, connectivity, transform):
-        yield s, v
+    yield from _shapes(source, mask, connectivity, transform)
 
 
 @ensure_env
@@ -154,7 +155,7 @@ def sieve(source, size, out=None, mask=None, connectivity=4):
 
     Returns
     -------
-    out : numpy ndarray
+    out : numpy.ndarray
         Result
 
     Notes
@@ -180,15 +181,21 @@ def sieve(source, size, out=None, mask=None, connectivity=4):
 
 @ensure_env
 def rasterize(
-        shapes,
-        out_shape=None,
-        fill=0,
-        out=None,
-        transform=IDENTITY,
-        all_touched=False,
-        merge_alg=MergeAlg.replace,
-        default_value=1,
-        dtype=None):
+    shapes,
+    out_shape=None,
+    fill=0,
+    nodata=None,
+    masked=False,
+    out=None,
+    transform=IDENTITY,
+    all_touched=False,
+    merge_alg=MergeAlg.replace,
+    default_value=1,
+    dtype=None,
+    skip_invalid=True,
+    dst_path=None,
+    dst_kwds=None,
+):
     """Return an image array with input geometries burned in.
 
     Warnings will be raised for any invalid or empty geometries, and
@@ -203,13 +210,18 @@ def rasterize(
         the `default_value` will be used. If `value` is `None` the
         `fill` value will be used.
     out_shape : tuple or list with 2 integers
-        Shape of output numpy ndarray.
+        Shape of output :class:`numpy.ndarray`.
     fill : int or float, optional
         Used as fill value for all areas not covered by input
         geometries.
-    out : numpy ndarray, optional
-        Array of same shape and data type as `source` in which to store
-        results.
+    nodata: float, optional
+        nodata value to use in output file or masked array.
+    masked: bool, optional. Default: False.
+        If True, return a masked array. Note: nodata is always set in
+        the case of file output.
+    out : numpy.ndarray, optional
+        Array in which to store results. If not provided, out_shape
+        and dtype are required.
     transform : Affine transformation object, optional
         Transformation from pixel coordinates of `source` to the
         coordinate system of the input `shapes`. See the `transform`
@@ -226,12 +238,20 @@ def rasterize(
                 the new value will be added to the existing raster.
     default_value : int or float, optional
         Used as value for all geometries, if not provided in `shapes`.
-    dtype : rasterio or numpy data type, optional
+    dtype : rasterio or numpy.dtype, optional
         Used as data type for results, if `out` is not provided.
+    skip_invalid : bool, optional
+        If True (default), invalid shapes will be skipped. If False,
+        ValueError will be raised.
+    dst_path : str or PathLike, optional
+        Path of output dataset
+    dst_kwds : dict, optional
+        Dictionary of creation options and other parameters that will be
+        overlaid on the profile of the output dataset.
 
     Returns
     -------
-    numpy ndarray
+    numpy.ndarray :
         If `out` was not None then `out` is returned, it will have been
         modified in-place. If `out` was None, this will be a new array.
 
@@ -252,45 +272,33 @@ def rasterize(
     shapes will be iterated multiple times. Performance is thus a linear
     function of buffer size. For maximum speed, ensure that
     GDAL_CACHEMAX is larger than the size of `out` or `out_shape`.
-
     """
     valid_dtypes = (
-        'int16', 'int32', 'uint8', 'uint16', 'uint32', 'float32', 'float64'
+        'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64', 'float32', 'float64'
     )
-    if GDALVersion.runtime().at_least("3.5"):
-        valid_dtypes += ("int64", "uint64")
     if GDALVersion.runtime().at_least("3.7"):
         valid_dtypes += ("int8",)
 
-    def format_invalid_dtype(param):
-        return '{0} dtype must be one of: {1}'.format(
-            param, ', '.join(valid_dtypes)
+    # The output data type is primarily determined by the output array
+    # or dtype parameter. But if neither of these are specified, it will
+    # be determined by the values that accompany the input shapes
+    # (below).
+    if out is not None:
+        dtype = out.dtype.name
+    elif dtype is not None:
+        dtype = np.dtype(dtype).name
+    else:
+        # dtype will be determined later.
+        pass
+
+    if dtype is not None and dtype not in valid_dtypes:
+        raise ValueError(
+            "Data type specified by out array or dtype parameter is not supported."
         )
 
-    def format_cast_error(param, dtype):
-        return '{0} cannot be cast to specified dtype: {1}'.format(param, dtype)
-
-    if fill != 0:
-        fill_array = np.array([fill])
-        if not validate_dtype(fill_array, valid_dtypes):
-            raise ValueError(format_invalid_dtype('fill'))
-
-        if dtype is not None and not can_cast_dtype(fill_array, dtype):
-            raise ValueError(format_cast_error('fill', dtype))
-
-    if default_value != 1:
-        default_value_array = np.array([default_value])
-        if not validate_dtype(default_value_array, valid_dtypes):
-            raise ValueError(format_invalid_dtype('default_value'))
-
-        if dtype is not None and not can_cast_dtype(default_value_array, dtype):
-            raise ValueError(format_cast_error('default_vaue', dtype))
-
-    if dtype is not None and _getnpdtype(dtype).name not in valid_dtypes:
-        raise ValueError(format_invalid_dtype('dtype'))
-
     valid_shapes = []
     shape_values = []
+
     for index, item in enumerate(shapes):
         if isinstance(item, (tuple, list)):
             geom, value = item
@@ -299,11 +307,11 @@ def rasterize(
         else:
             geom = item
             value = default_value
+
         geom = getattr(geom, '__geo_interface__', None) or geom
 
         if is_valid_geom(geom):
             shape_values.append(value)
-
             geom_type = geom['type']
 
             if geom_type == 'GeometryCollection':
@@ -324,47 +332,69 @@ def rasterize(
                 valid_shapes.append((geom, value))
 
         else:
-            # invalid or empty geometries are skipped and raise a warning instead
-            warnings.warn('Invalid or empty shape {} at index {} will not be rasterized.'.format(geom, index), ShapeSkipWarning)
+            if skip_invalid:
+                warnings.warn(
+                    "Invalid or empty shape {} at index {} will not be rasterized.".format(
+                        geom, index
+                    ),
+                    ShapeSkipWarning,
+                )
+            else:
+                raise ValueError("Invalid or empty shape cannot be rasterized.")
 
-    if not valid_shapes:
-        raise ValueError('No valid geometry objects found for rasterize')
+    # If neither an out array or dtype were given, we get the output
+    # data type from the shapes values, including the default.
+    if not dtype and valid_shapes:
+        values_arr = np.array(shape_values)
+        dtype = values_arr.dtype.name
 
-    shape_values = np.array(shape_values)
+        # GDAL 3.5 doesn't support int64 output. We'll try int32.
+        if dtype not in valid_dtypes and dtype.startswith("int"):
+            lo, hi = values_arr.min(), values_arr.max()
+            if -2147483648 <= lo and hi <= 2147483647:
+                dtype = "int32"
+            elif 0 <= lo and hi <= 4294967295:
+                dtype = "uint32"
+            else:
+                raise ValueError("GDAL versions < 3.6 cannot rasterize int64 values.")
 
-    if not validate_dtype(shape_values, valid_dtypes):
-        raise ValueError(format_invalid_dtype('shape values'))
+    with ExitStack() as exit_stack:
+        if dst_path is not None:
+            if isinstance(dst_path, DatasetWriter):
+                out = dst_path
+            else:
+                out_profile = dict(**dst_kwds)
+                if nodata is not None:
+                    out_profile["nodata"] = nodata
+                if dtype is not None:
+                    out_profile["dtype"] = dtype
+                out = rasterio.open(dst_path, "w", **out_profile)
+                exit_stack.enter_context(out)
 
-    if dtype is None:
-        dtype = get_minimum_dtype(np.append(shape_values, fill))
+        elif out is not None:
+            pass
 
-    elif not can_cast_dtype(shape_values, dtype):
-        raise ValueError(format_cast_error('shape values', dtype))
+        elif out_shape is not None:
+            if len(out_shape) != 2:
+                raise ValueError('Invalid out_shape, must be 2D')
+            out = np.empty(out_shape, dtype=dtype)
+            out.fill(fill)
 
-    if out is not None:
-        if _getnpdtype(out.dtype).name not in valid_dtypes:
-            raise ValueError(format_invalid_dtype('out'))
+        else:
+            raise ValueError('Either an out_shape or image must be provided')
 
-        if not can_cast_dtype(shape_values, out.dtype):
-            raise ValueError(format_cast_error('shape values', out.dtype.name))
+        if min(out.shape) == 0:
+            raise ValueError("width and height must be > 0")
 
-    elif out_shape is not None:
+        transform = guard_transform(transform)
 
-        if len(out_shape) != 2:
-            raise ValueError('Invalid out_shape, must be 2D')
+        if valid_shapes:
+            _rasterize(valid_shapes, out, transform, all_touched, merge_alg)
 
-        out = np.empty(out_shape, dtype=dtype)
-        out.fill(fill)
-
-    else:
-        raise ValueError('Either an out_shape or image must be provided')
-
-    if min(out.shape) == 0:
-        raise ValueError("width and height must be > 0")
-
-    transform = guard_transform(transform)
-    _rasterize(valid_shapes, out, transform, all_touched, merge_alg)
-    return out
+        if isinstance(out, np.ndarray):
+            if masked and not hasattr(out, "mask"):
+                out = np.ma.masked_equal(out, nodata, copy=False)
+            return out
 
 
 def bounds(geometry, north_up=True, transform=None):
@@ -547,7 +577,7 @@ def is_valid_geom(geom):
                     len(coords[0][0]) >= 2)
 
         if geom_type == 'MultiPolygon':
-            # Muti polygons must have at least one Polygon
+            # Multi polygons must have at least one Polygon
             return (len(coords) > 0 and len(coords[0]) > 0 and
                     len(coords[0][0]) >= 4 and len(coords[0][0][0]) >= 2)
 
@@ -692,12 +722,9 @@ def dataset_features(
                 antimeridian_cutting=True, precision=precision)
         xs, ys = zip(*coords(g))
         yield {
-            'type': 'Feature',
-            'id': "{0}:{1}".format(src_basename, i),
-            'properties': {
-                'val': val,
-                'filename': src_basename
-            },
-            'bbox': [min(xs), min(ys), max(xs), max(ys)],
-            'geometry': g
+            "type": "Feature",
+            "id": f"{src_basename}:{i}",
+            "properties": {"val": val, "filename": src_basename},
+            "bbox": [min(xs), min(ys), max(xs), max(ys)],
+            "geometry": g,
         }
