@@ -10,7 +10,7 @@ import rasterio
 from rasterio.coords import disjoint_bounds
 from rasterio.crs import CRS
 from rasterio.enums import MaskFlags
-from rasterio.windows import Window
+from rasterio import windows
 
 logger = logging.getLogger(__name__)
 
@@ -28,6 +28,11 @@ projection_projected_opt = click.option(
     default=True,
     help="Bounds in input's own projected coordinates (the default).")
 
+data_window_options = click.option(
+    "--to-data-window",
+    flag_value="datawindow",
+    help="Clip the raster to the region of valid data by removing areas of surrounding NoData values.",
+)
 
 # Clip command
 @click.command(short_help='Clip a raster to given bounds.')
@@ -39,10 +44,12 @@ projection_projected_opt = click.option(
     metavar="INPUT OUTPUT")
 @options.output_opt
 @options.bounds_opt
+@data_window_options
 @click.option(
     '--like',
     type=click.Path(exists=True),
-    help='Raster dataset to use as a template for bounds')
+    help='Raster dataset to use as a template for bounds'
+    )
 @options.format_opt
 @options.nodata_opt
 @projection_geographic_opt
@@ -67,6 +74,7 @@ def clip(
     overwrite,
     creation_options,
     with_complement,
+    to_data_window,
 ):
     """Clips a raster using projected or geographic bounds.
 
@@ -108,10 +116,12 @@ def clip(
                 if projection == 'geographic':
                     bounds = transform_bounds(CRS.from_epsg(4326), src.crs, *bounds)
                 if disjoint_bounds(bounds, src.bounds):
-                    raise click.BadParameter('must overlap the extent of '
-                                             'the input raster',
-                                             param='--bounds',
-                                             param_hint='--bounds')
+                    raise click.BadParameter(
+                        "must overlap the extent of " "the input raster",
+                        param="--bounds",
+                        param_hint="--bounds",
+                    )
+                bounds_window = src.window(*bounds)
             elif like:
                 with rasterio.open(like) as template_ds:
                     bounds = template_ds.bounds
@@ -120,24 +130,37 @@ def clip(
                                                   *bounds)
 
                     if disjoint_bounds(bounds, src.bounds):
-                        raise click.BadParameter('must overlap the extent of '
-                                                 'the input raster',
-                                                 param='--like',
-                                                 param_hint='--like')
+                        raise click.BadParameter(
+                            "must overlap the extent of " "the input raster",
+                            param="--like",
+                            param_hint="--like",
+                        )
+                bounds_window = src.window(*bounds)
+
+            elif to_data_window:
+                src_win = windows.Window(0, 0, width=src.width, height=src.height)
+                data_wins = []
+                for sw in windows.subdivide(src_win, 512, 512):
+                    dw = windows.get_data_window(src.read(1, masked=True, window=sw))
+                    if dw.width * dw.height > 0:
+                        # Adjust offsets
+                        dw = windows.Window(col_off=dw.col_off + sw.col_off,
+                                            row_off=dw.row_off + sw.row_off,
+                                            width=dw.width, height=dw.height)
+                        data_wins.append(dw)
+                bounds_window = windows.union(data_wins)
 
             else:
-                raise click.UsageError('--bounds or --like required')
-
-            bounds_window = src.window(*bounds)
+                raise click.UsageError("--bounds, --like, or --to-data-window required")
 
             if not with_complement:
                 bounds_window = bounds_window.intersection(
-                    Window(0, 0, src.width, src.height)
+                    windows.Window(0, 0, src.width, src.height)
                 )
 
-            # Get the window with integer height
-            # and width that contains the bounds window.
+            # Align window, as in gdal_translate.
             out_window = bounds_window.round_lengths()
+            out_window = out_window.round_offsets()
 
             height = int(out_window.height)
             width = int(out_window.width)
@@ -160,32 +183,37 @@ def clip(
             if "blockxsize" in out_kwargs and int(out_kwargs["blockxsize"]) > width:
                 del out_kwargs["blockxsize"]
                 logger.warning(
-                    "Blockxsize removed from creation options to accomodate small output width"
+                    "Blockxsize removed from creation options to accommodate small output width"
                 )
             if "blockysize" in out_kwargs and int(out_kwargs["blockysize"]) > height:
                 del out_kwargs["blockysize"]
                 logger.warning(
-                    "Blockysize removed from creation options to accomodate small output height"
+                    "Blockysize removed from creation options to accommodate small output height"
                 )
 
             with rasterio.open(output, "w", **out_kwargs) as out:
-                out.write(
-                    src.read(
-                        window=out_window,
-                        out_shape=(src.count, height, width),
-                        boundless=True,
-                        masked=True,
-                    )
-                )
-
-                if MaskFlags.per_dataset in src.mask_flag_enums[0]:
-                    out.write_mask(
-                        src.read_masks(
-                            window=out_window,
-                            out_shape=(src.count, height, width),
+                dest_win = windows.Window(0, 0, width, height)
+                for dw, sw in zip(windows.subdivide(dest_win, 512, 512), windows.subdivide(out_window, 512, 512)):
+                    out_shape = (src.count, sw.height, sw.width)
+                    out.write(
+                        src.read(
+                            window=sw,
+                            out_shape=out_shape,
                             boundless=True,
-                        )[0]
+                            masked=True,
+                        ),
+                        window=dw
                     )
 
-                # TODO: copy other properties (GCPs etc). Several other
-                # programs need the same utility.
+                    if MaskFlags.per_dataset in src.mask_flag_enums[0]:
+                        out.write_mask(
+                            src.read_masks(
+                                window=sw,
+                                out_shape=out_shape,
+                                boundless=True,
+                            )[0],
+                            window=dw
+                        )
+
+                    # TODO: copy other properties (GCPs etc). Several other
+                    # programs need the same utility.
