@@ -1,27 +1,23 @@
 """Geospatial transforms"""
 
-from collections.abc import Iterable
 from contextlib import ExitStack
 from functools import partial
-import math
 import numpy as np
-import sys
+import warnings
+from numbers import Number
 
 from affine import Affine
 
-import rasterio._loading
-with rasterio._loading.add_gdal_dll_directories():
-    import rasterio
-    from rasterio.env import env_ctx_if_needed
-    from rasterio._transform import (
-        _transform_from_gcps,
-        RPCTransformerBase,
-        GCPTransformerBase
-    )
-    from rasterio.enums import TransformDirection, TransformMethod
-    from rasterio.control import GroundControlPoint
-    from rasterio.rpc import RPC
-    from rasterio.errors import TransformError, RasterioDeprecationWarning
+from rasterio.env import env_ctx_if_needed
+from rasterio._transform import (
+    _transform_from_gcps,
+    RPCTransformerBase,
+    GCPTransformerBase,
+)
+from rasterio.enums import TransformDirection, TransformMethod
+from rasterio.control import GroundControlPoint
+from rasterio.rpc import RPC
+from rasterio.errors import TransformError, RasterioDeprecationWarning
 
 IDENTITY = Affine.identity()
 GDAL_IDENTITY = IDENTITY.to_gdal()
@@ -61,13 +57,13 @@ class TransformMethodsMixin:
         col : int
             Pixel column.
         z : float, optional
-            Height associated with coordinates. Primarily used for RPC based 
-            coordinate transformations. Ignored for affine based 
+            Height associated with coordinates. Primarily used for RPC based
+            coordinate transformations. Ignored for affine based
             transformations. Default: 0.
         offset : str, optional
             Determines if the returned coordinates are for the center of the
             pixel or for a corner.
-        transform_method: TransformMethod, optional 
+        transform_method: TransformMethod, optional
             The coordinate transformation method. Default: `TransformMethod.affine`.
         rpc_options: dict, optional
             Additional arguments passed to GDALCreateRPCTransformer
@@ -82,7 +78,7 @@ class TransformMethodsMixin:
         if transform_method is TransformMethod.gcps:
             transform = transform[0]
         if not transform:
-            raise AttributeError("Dataset has no {}".format(transform_method))
+            raise AttributeError(f"Dataset has no {transform_method}")
         return xy(transform, row, col, zs=z, offset=offset, **rpc_options)
 
     def index(
@@ -90,7 +86,7 @@ class TransformMethodsMixin:
         x,
         y,
         z=None,
-        op=math.floor,
+        op=None,
         precision=None,
         transform_method=TransformMethod.affine,
         **rpc_options
@@ -104,14 +100,15 @@ class TransformMethodsMixin:
         y : float
             y value in coordinate reference system
         z : float, optional
-            Height associated with coordinates. Primarily used for RPC based 
-            coordinate transformations. Ignored for affine based 
+            Height associated with coordinates. Primarily used for RPC
+            based coordinate transformations. Ignored for affine based
             transformations. Default: 0.
-        op : function, optional (default: math.floor)
-            Function to convert fractional pixels to whole numbers (floor,
-            ceiling, round)
-        transform_method: TransformMethod, optional 
-            The coordinate transformation method. Default: `TransformMethod.affine`.
+        op : function, optional (default: numpy.floor)
+            Function to convert fractional pixels to whole numbers
+            (floor, ceiling, round)
+        transform_method: TransformMethod, optional
+            The coordinate transformation method. Default:
+            `TransformMethod.affine`.
         rpc_options: dict, optional
             Additional arguments passed to GDALCreateRPCTransformer
         precision : int, optional
@@ -120,7 +117,7 @@ class TransformMethodsMixin:
 
         Returns
         -------
-        tuple
+        tuple: int, int
             (row index, col index)
 
         """
@@ -134,8 +131,8 @@ class TransformMethodsMixin:
         if transform_method is TransformMethod.gcps:
             transform = transform[0]
         if not transform:
-            raise AttributeError("Dataset has no {}".format(transform_method))
-        return rowcol(transform, x, y, zs=z, op=op, **rpc_options)
+            raise AttributeError(f"Dataset has no {transform_method}")
+        return tuple(int(val) for val in rowcol(transform, x, y, zs=z, op=op, **rpc_options))
 
 
 def get_transformer(transform, **rpc_options):
@@ -200,9 +197,19 @@ def array_bounds(height, width, transform):
     its height, width, and an affine transform.
 
     """
-    w, n = transform.xoff, transform.yoff
-    e, s = transform * (width, height)
-    return w, s, e, n
+    a, b, c, d, e, f, _, _, _ = transform
+    if b == d == 0:
+        west, south, east, north = c, f + e * height, c + a * width, f
+    else:
+        c0x, c0y = c, f
+        c1x, c1y = transform * (0, height)
+        c2x, c2y = transform * (width, height)
+        c3x, c3y = transform * (width, 0)
+        xs = (c0x, c1x, c2x, c3x)
+        ys = (c0y, c1y, c2y, c3y)
+        west, south, east, north = min(xs), min(ys), max(xs), max(ys)
+
+    return west, south, east, north
 
 
 def xy(transform, rows, cols, zs=None, offset='center', **rpc_options):
@@ -211,20 +218,20 @@ def xy(transform, rows, cols, zs=None, offset='center', **rpc_options):
     The pixel's center is returned by default, but a corner can be returned
     by setting `offset` to one of `ul, ur, ll, lr`.
 
-    Supports affine, Ground Control Point (GCP), or Rational Polynomial 
+    Supports affine, Ground Control Point (GCP), or Rational Polynomial
     Coefficients (RPC) based coordinate transformations.
 
     Parameters
     ----------
     transform : Affine or sequence of GroundControlPoint or RPC
-        Transform suitable for input to AffineTransformer, GCPTransformer, or RPCTransformer. 
+        Transform suitable for input to AffineTransformer, GCPTransformer, or RPCTransformer.
     rows : list or int
         Pixel rows.
     cols : int or sequence of ints
         Pixel columns.
     zs : list or float, optional
-        Height associated with coordinates. Primarily used for RPC based 
-        coordinate transformations. Ignored for affine based 
+        Height associated with coordinates. Primarily used for RPC based
+        coordinate transformations. Ignored for affine based
         transformations. Default: 0.
     offset : str, optional
         Determines if the returned coordinates are for the center of the
@@ -245,24 +252,33 @@ def xy(transform, rows, cols, zs=None, offset='center', **rpc_options):
         return transformer.xy(rows, cols, zs=zs, offset=offset)
 
 
-def rowcol(transform, xs, ys, zs=None, op=math.floor, precision=None, **rpc_options):
+def rowcol(
+    transform,
+    xs,
+    ys,
+    zs=None,
+    op=None,
+    precision=None,
+    **rpc_options,
+):
     """Get rows and cols of the pixels containing (x, y).
 
     Parameters
     ----------
     transform : Affine or sequence of GroundControlPoint or RPC
-        Transform suitable for input to AffineTransformer, GCPTransformer, or RPCTransformer.
+        Transform suitable for input to AffineTransformer,
+        GCPTransformer, or RPCTransformer.
     xs : list or float
         x values in coordinate reference system.
     ys : list or float
         y values in coordinate reference system.
     zs : list or float, optional
-        Height associated with coordinates. Primarily used for RPC based 
-        coordinate transformations. Ignored for affine based 
+        Height associated with coordinates. Primarily used for RPC based
+        coordinate transformations. Ignored for affine based
         transformations. Default: 0.
-    op : function
-        Function to convert fractional pixels to whole numbers (floor, ceiling,
-        round).
+    op : function, optional (default: numpy.floor)
+        Function to convert fractional pixels to whole numbers (floor,
+        ceiling, round)
     precision : int or float, optional
         This parameter is unused, deprecated in rasterio 1.3.0, and
         will be removed in version 2.0.0.
@@ -271,10 +287,10 @@ def rowcol(transform, xs, ys, zs=None, op=math.floor, precision=None, **rpc_opti
 
     Returns
     -------
-    rows : list of ints
-        list of row indices
-    cols : list of ints
-        list of column indices
+    rows : array of ints or floats
+    cols : array of ints or floats
+        Integers are the default. The numerical type is determined by
+        the type returned by op().
 
     """
     if precision is not None:
@@ -304,7 +320,7 @@ def from_gcps(gcps):
     return Affine.from_gdal(*_transform_from_gcps(gcps))
 
 
-class TransformerBase():
+class TransformerBase:
     """Generic GDAL transformer base class
 
     Notes
@@ -313,39 +329,38 @@ class TransformerBase():
 
     """
     def __init__(self):
-        self._env = ExitStack()
         self._transformer = None
-        self.closed = True
-
-    def close(self):
-        self.closed = True
 
     @staticmethod
     def _ensure_arr_input(xs, ys, zs=None):
         """Ensure all input coordinates are mapped to array-like objects
-        
+
         Raises
         ------
         TransformError
             If input coordinates are not all of the same length
         """
+        xs = np.atleast_1d(xs)
+        ys = np.atleast_1d(ys)
+        if zs is not None:
+            zs = np.atleast_1d(zs)
+        else:
+            zs = np.zeros(1)
+
         try:
-            xs, ys, zs = np.broadcast_arrays(xs, ys, 0 if zs is None else zs)
+            broadcasted = np.broadcast(xs, ys, zs)
         except ValueError as error:
-            raise TransformError(
-                "Input coordinates must be broadcastable to a 1d array"
-            ) from error
-        return np.atleast_1d(xs), np.atleast_1d(ys), np.atleast_1d(zs)
+            raise TransformError() from error
+
+        return xs, ys, zs
 
     def __enter__(self):
-        self._env.enter_context(env_ctx_if_needed())
         return self
 
     def __exit__(self, *args):
-        self.close()
-        self._env.close()
-    
-    def rowcol(self, xs, ys, zs=None, op=math.floor, precision=None):
+        pass
+
+    def rowcol(self, xs, ys, zs=None, op=None, precision=None):
         """Get rows and cols coordinates given geographic coordinates.
 
         Parameters
@@ -353,24 +368,28 @@ class TransformerBase():
         xs, ys : float or list of float
             Geographic coordinates
         zs : float or list of float, optional
-            Height associated with coordinates. Primarily used for RPC based 
-            coordinate transformations. Ignored for affine based 
+            Height associated with coordinates. Primarily used for RPC
+            based coordinate transformations. Ignored for affine based
             transformations. Default: 0.
-        op : function, optional (default: math.floor)
-            Function to convert fractional pixels to whole numbers (floor,
-            ceiling, round)
+        op : function, optional (default: numpy.floor)
+            Function to convert fractional pixels to whole numbers
+            (floor, ceiling, round)
         precision : int, optional (default: None)
             This parameter is unused, deprecated in rasterio 1.3.0, and
             will be removed in version 2.0.0.
 
         Raises
         ------
+        TypeError
+            If coordinate transformation fails.
         ValueError
-            If input coordinates are not all equal length
+            If input coordinates are not all equal length.
 
         Returns
         -------
-        tuple of float or list of float.
+        tuple of numbers or array of numbers.
+            Integers are the default. The numerical type is determined
+            by the type returned by op().
 
         """
         if precision is not None:
@@ -379,18 +398,29 @@ class TransformerBase():
                 RasterioDeprecationWarning,
             )
 
-        AS_ARR = True if hasattr(xs, "__iter__") else False
+        IS_SCALAR = isinstance(xs, Number) and isinstance(ys, Number)
         xs, ys, zs = self._ensure_arr_input(xs, ys, zs=zs)
-        
+
         try:
             new_cols, new_rows = self._transform(
                 xs, ys, zs, transform_direction=TransformDirection.reverse
             )
 
-            if len(new_rows) == 1 and not AS_ARR:
-                return (op(new_rows[0]), op(new_cols[0]))
+            if op is None:
+                new_rows = np.floor(new_rows).astype(dtype="int32")
+                new_cols = np.floor(new_cols).astype(dtype="int32")
+            elif isinstance(op, np.ufunc):
+                op(new_rows, out=new_rows)
+                op(new_cols, out=new_cols)
             else:
-                return ([op(r) for r in new_rows], [op(c) for c in new_cols])
+                new_rows = np.array(list(map(op, new_rows)))
+                new_cols = np.array(list(map(op, new_cols)))
+
+            if IS_SCALAR:
+                return new_rows[0], new_cols[0]
+            else:
+                return new_rows, new_cols
+
         except TypeError:
             raise TransformError("Invalid inputs")
 
@@ -403,12 +433,13 @@ class TransformerBase():
         rows, cols : int or list of int
             Image pixel coordinates
         zs : float or list of float, optional
-            Height associated with coordinates. Primarily used for RPC based 
-            coordinate transformations. Ignored for affine based 
+            Height associated with coordinates. Primarily used for RPC based
+            coordinate transformations. Ignored for affine based
             transformations. Default: 0.
         offset : str, optional
             Determines if the returned coordinates are for the center of the
-            pixel or for a corner.
+            pixel or for a corner. Available options include center, ul, ur, ll,
+            lr.
         Raises
         ------
         ValueError
@@ -419,9 +450,9 @@ class TransformerBase():
         tuple of float or list of float
 
         """
-        AS_ARR = True if hasattr(rows, "__iter__") else False
+        IS_SCALAR = isinstance(rows, Number) and isinstance(cols, Number)
         rows, cols, zs = self._ensure_arr_input(rows, cols, zs=zs)
-        
+
         if offset == 'center':
             coff, roff = (0.5, 0.5)
         elif offset == 'ul':
@@ -434,32 +465,46 @@ class TransformerBase():
             coff, roff = (1, 1)
         else:
             raise TransformError("Invalid offset")
-        
-        # shift input coordinates according to offset
-        T = IDENTITY.translation(coff, roff)
-        temp_rows = []
-        temp_cols = []
-        try:
-            for pt in zip(cols, rows):
-                y, x = T * pt
-                temp_rows.append(y)
-                temp_cols.append(x)
 
-            new_ys, new_xs = self._transform(
-                temp_rows, temp_cols, zs, transform_direction=TransformDirection.forward
+        try:
+            # shift input coordinates according to offset
+            T = IDENTITY.translation(coff, roff)
+            identity_transformer = AffineTransformer(T)
+            offset_cols, offset_rows = identity_transformer._transform(
+                cols, rows, zs, transform_direction=TransformDirection.forward
+            )
+            new_xs, new_ys = self._transform(
+                offset_cols, offset_rows, zs, transform_direction=TransformDirection.forward
             )
 
-            if len(new_ys) == 1 and not AS_ARR:
-                return (new_ys[0], new_xs[0])
+            if IS_SCALAR:
+                return new_xs[0], new_ys[0]
             else:
-                return (new_ys, new_xs)
+                return new_xs, new_ys
         except TypeError:
             raise TransformError("Invalid inputs")
-    
+
     def _transform(self, xs, ys, zs, transform_direction):
         raise NotImplementedError
 
 
+class GDALTransformerBase(TransformerBase):
+    def __init__(self):
+        super().__init__()
+        self._env = ExitStack()
+
+    def close(self):
+        pass
+
+    def __enter__(self):
+        self._env.enter_context(env_ctx_if_needed())
+        return self
+
+    def __exit__(self, *args):
+        self.close()
+        self._env.close()
+
+
 class AffineTransformer(TransformerBase):
     """A pure Python class related to affine based coordinate transformations."""
     def __init__(self, affine_transform):
@@ -467,38 +512,36 @@ class AffineTransformer(TransformerBase):
         if not isinstance(affine_transform, Affine):
             raise ValueError("Not an affine transform")
         self._transformer = affine_transform
+        self._transform_arr = np.empty((3, 3), dtype='float64')
 
-    def close(self):
-        pass
-    
     def _transform(self, xs, ys, zs, transform_direction):
-        resxs = []
-        resys = []
-        
+        transform = self._transform_arr
         if transform_direction is TransformDirection.forward:
-            transform = self._transformer
+            transform.flat[:] = self._transformer
         elif transform_direction is TransformDirection.reverse:
-            transform = ~self._transformer
+            transform.flat[:] = ~self._transformer
+
+        bi = np.broadcast(xs, ys)
+        input_matrix = np.empty((3, bi.size))
+        input_matrix[0] = bi.iters[0]
+        input_matrix[1] = bi.iters[1]
+        input_matrix[2] = 1
+        transform.dot(input_matrix, out=input_matrix)
+        return input_matrix[0], input_matrix[1]
+
 
-        for x, y in zip(xs, ys):
-            resx, resy = transform * (x, y)
-            resxs.append(resx)
-            resys.append(resy)
-        
-        return (resxs, resys)
-    
     def __repr__(self):
         return "<AffineTransformer>"
 
 
-class RPCTransformer(RPCTransformerBase, TransformerBase):
+class RPCTransformer(RPCTransformerBase, GDALTransformerBase):
     """
     Class related to Rational Polynomial Coeffecients (RPCs) based
     coordinate transformations.
 
     Uses GDALCreateRPCTransformer and GDALRPCTransform for computations. Options
     for GDALCreateRPCTransformer may be passed using `rpc_options`.
-    Ensure that GDAL transformer objects are destroyed by calling `close()` 
+    Ensure that GDAL transformer objects are destroyed by calling `close()`
     method or using context manager interface.
 
     """
@@ -512,20 +555,21 @@ class RPCTransformer(RPCTransformerBase, TransformerBase):
             self.closed and 'closed' or 'open')
 
 
-class GCPTransformer(GCPTransformerBase, TransformerBase):
+class GCPTransformer(GCPTransformerBase, GDALTransformerBase):
     """
     Class related to Ground Control Point (GCPs) based
     coordinate transformations.
 
     Uses GDALCreateGCPTransformer and GDALGCPTransform for computations.
-    Ensure that GDAL transformer objects are destroyed by calling `close()` 
-    method or using context manager interface.
+    Ensure that GDAL transformer objects are destroyed by calling `close()`
+    method or using context manager interface. If `tps` is set to True,
+    uses GDALCreateTPSTransformer and GDALTPSTransform instead.
 
     """
-    def __init__(self, gcps):
+    def __init__(self, gcps, tps=False):
         if len(gcps) and not isinstance(gcps[0], GroundControlPoint):
             raise ValueError("GCPTransformer requires sequence of GroundControlPoint")
-        super().__init__(gcps)
+        super().__init__(gcps, tps)
 
     def __repr__(self):
         return "<{} GCPTransformer>".format(
