@@ -528,8 +528,7 @@ def validate_length_value(instance, attribute, value):
         raise ValueError("Number of columns or rows must be non-negative")
 
 
-@attr.s(slots=True,
-        frozen=True)
+@attr.s(slots=True, frozen=True)
 class Window:
     """Windows are rectangular subsets of rasters.
 
@@ -722,9 +721,8 @@ class Window:
         Window
 
         """
-        operator = lambda x: int(math.floor(x + 0.5))
-        width = operator(self.width)
-        height = operator(self.height)
+        width = math.floor(self.width + 0.5)
+        height = math.floor(self.height + 0.5)
         return Window(self.col_off, self.row_off, width, height)
 
     def round_shape(self, **kwds):
@@ -750,11 +748,23 @@ class Window:
         Window
 
         """
-        operator = lambda x: int(math.floor(x + 0.001))
-        row_off = operator(self.row_off)
-        col_off = operator(self.col_off)
+        row_off = math.floor(self.row_off + 0.1)
+        col_off = math.floor(self.col_off + 0.1)
         return Window(col_off, row_off, self.width, self.height)
 
+    def round(self, ndigits=None):
+        """Round a window's offsets and lengths
+
+        Rounding to a very small fraction of a pixel can help treat
+        floating point issues arising from computation of windows.
+        """
+        return Window(
+            round(self.col_off, ndigits=ndigits),
+            round(self.row_off, ndigits=ndigits),
+            round(self.width, ndigits=ndigits),
+            round(self.height, ndigits=ndigits),
+        )
+
     def crop(self, height, width):
         """Return a copy cropped to height and width"""
         return crop(self, height, width)
@@ -773,3 +783,50 @@ class Window:
         Window
         """
         return intersection([self, other])
+
+
+def subdivide(window, height, width):
+    """Divide a window into smaller windows.
+
+    Windows have no overlap and will be at most the desired
+    height and width. Smaller windows will be generated where
+    the height and width do not evenly divide the window dimensions.
+
+    Parameters
+    ----------
+    window : Window
+        Source window to subdivide.
+    height : int
+        Subwindow height.
+    width : int
+        Subwindow width.
+
+    Returns
+    -------
+    list of Windows
+    """
+    subwindows = []
+
+    irow = window.row_off + window.height
+    icol = window.col_off + window.width
+
+    row_off = window.row_off
+    col_off = window.col_off
+    while row_off < irow:
+        if row_off + height > irow:
+            _height = irow - row_off
+        else:
+            _height = height
+
+        while col_off < icol:
+            if col_off + width > icol:
+                _width = icol - col_off
+            else:
+                _width = width
+
+            subwindows.append(Window(col_off, row_off, _width, _height))
+            col_off += width
+
+        row_off += height
+        col_off = window.col_off
+    return subwindows
